@@ -14,15 +14,13 @@
 # limitations under the License.
 #===============================================================================
 
-"""
-Contains functions and classes related to fields.
+""" Contains functions and classes related to fields.
 """
 
-import datetime
-import re
+import datetime, re, struct
 
-from whoosh.analysis import IDAnalyzer, RegexAnalyzer, KeywordAnalyzer
-from whoosh.analysis import StandardAnalyzer, NgramAnalyzer
+from whoosh.analysis import (IDAnalyzer, RegexAnalyzer, KeywordAnalyzer,
+                             StandardAnalyzer, NgramAnalyzer)
 from whoosh.formats import Format, Existence, Frequency, Positions
 
 # Exceptions
@@ -33,7 +31,6 @@ class UnknownFieldError(Exception):
     pass
 
 
-
 # Field Types
 
 class FieldType(object):
@@ -60,20 +57,21 @@ class FieldType(object):
       fields marked as 'unique' to find the previous version of a document
       being updated.
       
-    The constructor for the base field type simply lets you supply your
-    own configured field format, vector format, and scorable and stored
-    values. Subclasses may configure some or all of this for you.
+    The constructor for the base field type simply lets you supply your own
+    configured field format, vector format, and scorable and stored values.
+    Subclasses may configure some or all of this for you.
     
     """
     
     format = vector = scorable = stored = unique = None
+    parse_query = None
     indexed = True
     __inittypes__ = dict(format=Format, vector=Format,
                          scorable=bool, stored=bool, unique=bool)
     
-    def __init__(self, format, vector = None,
-                 scorable = False, stored = False,
-                 unique = False):
+    def __init__(self, format, vector=None,
+                 scorable=False, stored=False,
+                 unique=False):
         self.format = format
         self.vector = vector
         self.scorable = scorable
@@ -81,9 +79,9 @@ class FieldType(object):
         self.unique = unique
     
     def __repr__(self):
-        return "%s(format=%r, vector=%r, scorable=%s, stored=%s, unique=%s)"\
-        % (self.__class__.__name__, self.format, self.vector,
-           self.scorable, self.stored, self.unique)
+        temp = "%s(format=%r, vector=%r, scorable=%s, stored=%s, unique=%s)"
+        return temp % (self.__class__.__name__, self.format, self.vector,
+                       self.scorable, self.stored, self.unique)
     
     def __eq__(self, other):
         return all((isinstance(other, FieldType),
@@ -94,7 +92,8 @@ class FieldType(object):
                     (self.unique == other.unique)))
     
     def clean(self):
-        """Clears any cached information in the field and any child objects."""
+        """Clears any cached information in the field and any child objects.
+        """
         
         if self.format and hasattr(self.format, "clean"):
             self.format.clean()
@@ -107,26 +106,30 @@ class FieldType(object):
         
         if not self.format:
             raise Exception("%s field cannot index without a format" % self.__class__)
-        
         if not isinstance(value, unicode):
             raise ValueError("%r is not unicode" % value)
-        
         return self.format.word_values(value, mode="index", **kwargs)
     
+    def process_text(self, qstring, mode='', **kwargs):
+        if not self.format:
+            raise Exception("%s field has no format" % self)
+        return (t.text for t
+                in self.format.analyze(qstring, mode=mode, **kwargs))
+    
 
 class ID(FieldType):
     """Configured field type that indexes the entire value of the field as one
-    token. This is useful for data you don't want to tokenize, such as the
-    path of a file.
+    token. This is useful for data you don't want to tokenize, such as the path
+    of a file.
     """
     
     __inittypes__ = dict(stored=bool, unique=bool, field_boost=float)
     
-    def __init__(self, stored = False, unique = False, field_boost = 1.0):
+    def __init__(self, stored=False, unique=False, field_boost=1.0):
         """
         :param stored: Whether the value of this field is stored with the document.
         """
-        self.format = Existence(analyzer = IDAnalyzer(), field_boost = field_boost)
+        self.format = Existence(analyzer=IDAnalyzer(), field_boost=field_boost)
         self.stored = stored
         self.unique = unique
 
@@ -138,28 +141,89 @@ class IDLIST(FieldType):
     
     __inittypes__ = dict(stored=bool, unique=bool, expression=bool, field_boost=float)
     
-    def __init__(self, stored = False, unique = False, expression = None, field_boost = 1.0):
+    def __init__(self, stored=False, unique=False, expression=None, field_boost=1.0):
         """
-        :param stored: Whether the value of this field is stored with the document.
+        :param stored: Whether the value of this field is stored with the
+            document.
         :param unique: Whether the value of this field is unique per-document.
-        :param expression: The regular expression object to use to extract tokens.
-            The default expression breaks tokens on CRs, LFs, tabs, spaces, commas,
-            and semicolons.
+        :param expression: The regular expression object to use to extract
+            tokens. The default expression breaks tokens on CRs, LFs, tabs,
+            spaces, commas, and semicolons.
         """
         
         expression = expression or re.compile(r"[^\r\n\t ,;]+")
-        analyzer = RegexAnalyzer(expression = expression)
-        self.format = Existence(analyzer = analyzer, field_boost = field_boost)
+        analyzer = RegexAnalyzer(expression=expression)
+        self.format = Existence(analyzer=analyzer, field_boost=field_boost)
         self.stored = stored
         self.unique = unique
 
 
+class NUMERIC(FieldType):
+    def __init__(self, type=int, stored=False, unique=False, field_boost=1.0):
+        self.type = type
+        self.stored = stored
+        self.unique = unique
+        self.format = Existence(analyzer=IDAnalyzer(), field_boost=field_boost)
+    
+    def index(self, num):
+        method = getattr(self, self.type.__name__ + "_to_text")
+        return [(method(num), 1, '')]
+    
+    def to_text(self, x):
+        ntype = self.type
+        method = getattr(self, ntype.__name__ + "_to_text")
+        return method(ntype(x))
+    
+    def process_text(self, text, **kwargs):
+        return (self.to_text(text),)
+    
+    def parse_query(self, fieldname, qstring, boost=1.0):
+        from whoosh import query
+        return query.Term(fieldname, self.to_text(qstring), boost=boost)
+    
+    @staticmethod
+    def int_to_text(x):
+        x += (1 << (4 << 2)) - 1 # 4 means 32-bits
+        return u"%08x" % x
+    
+    @staticmethod
+    def text_to_int(text):
+        x = int(text, 16)
+        x -= (1 << (4 << 2)) - 1
+        return x
+    
+    @staticmethod
+    def long_to_text(x):
+        x += (1 << (8 << 2)) - 1
+        return u"%016x" % x
+    
+    @staticmethod
+    def text_to_long(text):
+        x = long(text, 16)
+        x -= (1 << (8 << 2)) - 1
+        return x
+    
+    @staticmethod
+    def float_to_text(x):
+        x = struct.unpack("<q", struct.pack("<d", x))[0]
+        x += (1 << (8 << 2)) - 1
+        return u"%016x" % x
+    
+    @staticmethod
+    def text_to_float(text):
+        x = long(text, 16)
+        x -= (1 << (8 << 2)) - 1
+        x = struct.unpack("<d", struct.pack("<q", x))[0]
+        return x
+    
+
 class DATETIME(FieldType):
     __inittypes__ = dict(stored=bool, unique=bool)
     
-    def __init__(self, stored = True, unique = False):
+    def __init__(self, stored=False, unique=False):
         """
-        :param stored: Whether the value of this field is stored with the document.
+        :param stored: Whether the value of this field is stored with the
+            document.
         :param unique: Whether the value of this field is unique per-document.
         """
         
@@ -171,10 +235,48 @@ class DATETIME(FieldType):
         if not isinstance(dt, datetime.datetime):
             raise ValueError("Value of DATETIME field must be a datetime object: %r" % dt)
         
-        text = dt.isoformat()
+        text = dt.isoformat() # 2010-02-02T17:06:19.109000
         text = text.replace(" ", "").replace(":", "").replace("-", "").replace(".", "")
         return [(text, 1, '')]
     
+    def process_text(self, text, **kwargs):
+        text = text.replace(" ", "").replace(":", "").replace("-", "").replace(".", "")
+        return (text,)
+    
+    def parse_query(self, fieldname, qstring, boost=1.0):
+        text = self.process_text(qstring)
+        from whoosh import query
+        return query.Prefix(fieldname, text, boost=boost)
+    
+
+class BOOLEAN(FieldType):
+    strings = (u"t", u"f")
+    trues = frozenset((u"t", u"true", u"yes", u"1"))
+    falses = frozenset((u"f", u"false", u"no", u"0"))
+    
+    __inittypes__ = dict(stored=bool)
+    
+    def __init__(self, stored=False):
+        self.stored = stored
+        self.format = Existence()
+    
+    def index(self, bit):
+        if not isinstance(bit, bool):
+            raise ValueError("Value of BOOL field must be a bool object: %r" % bit)
+        return [(self.strings[int(bit)], 1, '')]
+    
+    def parse_query(self, fieldname, qstring, boost=1.0):
+        from whoosh import query
+        text = None
+        if qstring in self.falses:
+            text = self.strings[0]
+        elif qstring in self.trues:
+            text = self.strings[1]
+        
+        if text is None:
+            return query.NullQuery
+        return query.Term(fieldname, text, boost=boost)
+    
 
 class STORED(FieldType):
     """Configured field type for fields you want to store but not index.
@@ -188,50 +290,53 @@ class STORED(FieldType):
     
 
 class KEYWORD(FieldType):
-    """Configured field type for fields containing space-separated or comma-separated
-    keyword-like data (such as tags). The default is to not store positional information
-    (so phrase searching is not allowed in this field) and to not make the field scorable.
+    """Configured field type for fields containing space-separated or
+    comma-separated keyword-like data (such as tags). The default is to not
+    store positional information (so phrase searching is not allowed in this
+    field) and to not make the field scorable.
     """
     
     __inittypes__ = dict(stored=bool, lowercase=bool, commas=bool, scorable=bool,
                          unique=bool, field_boost=float)
     
-    def __init__(self, stored = False, lowercase = False, commas = False,
-                 scorable = False, unique = False, field_boost = 1.0):
+    def __init__(self, stored=False, lowercase=False, commas=False,
+                 scorable=False, unique=False, field_boost=1.0):
         """
-        :param stored: Whether to store the value of the field with the document.
+        :param stored: Whether to store the value of the field with the
+            document.
         :param comma: Whether this is a comma-separated field. If this is False
             (the default), it is treated as a space-separated field.
         :param scorable: Whether this field is scorable.
         """
         
-        ana = KeywordAnalyzer(lowercase = lowercase, commas = commas)
-        self.format = Frequency(analyzer = ana, field_boost = field_boost)
+        ana = KeywordAnalyzer(lowercase=lowercase, commas=commas)
+        self.format = Frequency(analyzer=ana, field_boost=field_boost)
         self.scorable = scorable
         self.stored = stored
         self.unique = unique
 
 
 class TEXT(FieldType):
-    """Configured field type for text fields (for example, the body text of an article). The
-    default is to store positional information to allow phrase searching. This field type
-    is always scorable.
+    """Configured field type for text fields (for example, the body text of an
+    article). The default is to store positional information to allow phrase
+    searching. This field type is always scorable.
     """
     
     __inittypes__ = dict(analyzer=object, phrase=bool, vector=Format,
                          stored=bool, field_boost=float)
     
-    def __init__(self, analyzer = None, phrase = True, vector = None,
-                 stored = False, field_boost = 1.0):
+    def __init__(self, analyzer=None, phrase=True, vector=None,
+                 stored=False, field_boost=1.0):
         """
-        :param stored: Whether to store the value of this field with the document. Since
-            this field type generally contains a lot of text, you should avoid storing it
-            with the document unless you need to, for example to allow fast excerpts in the
-            search results.
-        :param phrase: Whether the store positional information to allow phrase searching.
-        :param analyzer: The analysis.Analyzer to use to index the field contents. See the
-            analysis module for more information. If you omit this argument, the field uses
-            analysis.StandardAnalyzer.
+        :param stored: Whether to store the value of this field with the
+            document. Since this field type generally contains a lot of text,
+            you should avoid storing it with the document unless you need to,
+            for example to allow fast excerpts in the search results.
+        :param phrase: Whether the store positional information to allow phrase
+            searching.
+        :param analyzer: The analysis.Analyzer to use to index the field
+            contents. See the analysis module for more information. If you omit
+            this argument, the field uses analysis.StandardAnalyzer.
         """
         
         ana = analyzer or StandardAnalyzer()
@@ -240,7 +345,7 @@ class TEXT(FieldType):
             formatclass = Positions
         else:
             formatclass = Frequency
-        self.format = formatclass(analyzer = ana, field_boost = field_boost)
+        self.format = formatclass(analyzer=ana, field_boost=field_boost)
         self.vector = vector
         
         self.scorable = True
@@ -248,25 +353,25 @@ class TEXT(FieldType):
 
 
 class NGRAM(FieldType):
-    """Configured field that indexes text as N-grams. For example, with a field type
-    NGRAM(3,4), the value "hello" will be indexed as tokens
+    """Configured field that indexes text as N-grams. For example, with a field
+    type NGRAM(3,4), the value "hello" will be indexed as tokens
     "hel", "hell", "ell", "ello", "llo".
     """
     
     __inittypes__ = dict(minsize=int, maxsize=int, stored=bool, field_boost=float)
     
-    def __init__(self, minsize = 2, maxsize = 4, stored = False, field_boost = 1.0):
+    def __init__(self, minsize=2, maxsize=4, stored=False, field_boost=1.0):
         """
-        :param stored: Whether to store the value of this field with the document. Since
-            this field type generally contains a lot of text, you should avoid storing it
-            with the document unless you need to, for example to allow fast excerpts in the
-            search results.
+        :param stored: Whether to store the value of this field with the
+            document. Since this field type generally contains a lot of text,
+            you should avoid storing it with the document unless you need to,
+            for example to allow fast excerpts in the search results.
         :param minsize: The minimum length of the N-grams.
         :param maxsize: The maximum length of the N-grams.
         """
         
-        self.format = Frequency(analyzer = NgramAnalyzer(minsize, maxsize),
-                                field_boost = field_boost)
+        self.format = Frequency(analyzer=NgramAnalyzer(minsize, maxsize),
+                                field_boost=field_boost)
         self.scorable = True
         self.stored = stored
 
@@ -277,17 +382,16 @@ class Schema(object):
     """Represents the collection of fields in an index. Maps field names to
     FieldType objects which define the behavior of each field.
     
-    Low-level parts of the index use field numbers instead of field names
-    for compactness. This class has several methods for converting between
-    the field name, field number, and field object itself.
+    Low-level parts of the index use field numbers instead of field names for
+    compactness. This class has several methods for converting between the
+    field name, field number, and field object itself.
     """
     
     def __init__(self, **fields):
-        """
-        All keyword arguments to the constructor are treated as fieldname = fieldtype
-        pairs. The fieldtype can be an instantiated FieldType object, or a FieldType
-        sub-class (in which case the Schema will instantiate it with the default
-        constructor before adding it).
+        """ All keyword arguments to the constructor are treated as fieldname =
+        fieldtype pairs. The fieldtype can be an instantiated FieldType object,
+        or a FieldType sub-class (in which case the Schema will instantiate it
+        with the default constructor before adding it).
         
         For example::
         
@@ -312,15 +416,13 @@ class Schema(object):
         return "<Schema: %s>" % repr(self._names)
     
     def __iter__(self):
-        """
-        Yields the sequence of fields in this schema.
+        """Yields the sequence of fields in this schema.
         """
         
         return iter(self._by_number)
     
     def __getitem__(self, id):
-        """
-        Returns the field associated with the given field name or number.
+        """Returns the field associated with the given field name or number.
         
         :param id: A field name or field number.
         """
@@ -330,58 +432,56 @@ class Schema(object):
         return self._by_number[id]
     
     def __len__(self):
-        """
-        Returns the number of fields in this schema.
+        """Returns the number of fields in this schema.
         """
         return len(self._by_number)
     
     def __contains__(self, fieldname):
-        """
-        Returns True if a field by the given name is in this schema.
+        """Returns True if a field by the given name is in this schema.
         
         :param fieldname: The name of the field.
         """
         return fieldname in self._by_name
     
+    def copy(self):
+        import copy
+        return copy.deepcopy(self)
+    
     def field_by_name(self, name):
-        """
-        Returns the field object associated with the given name.
+        """Returns the field object associated with the given name.
         
         :param name: The name of the field to retrieve.
         """
         return self._by_name[name]
     
     def field_by_number(self, number):
-        """
-        Returns the field object associated with the given number.
+        """Returns the field object associated with the given number.
         
         :param number: The number of the field to retrieve.
         """
         return self._by_number[number]
     
     def fields(self):
-        """
-        Yields ("fieldname", field_object) pairs for the fields
-        in this schema.
+        """Yields ("fieldname", field_object) pairs for the fields in this
+        schema.
         """
         return self._by_name.iteritems()
     
     def field_names(self):
-        """
-        Returns a list of the names of the fields in this schema.
+        """Returns a list of the names of the fields in this schema.
         """
         return self._names
     
     def add(self, name, fieldtype):
-        """
-        Adds a field to this schema. This is a low-level method; use keyword
+        """Adds a field to this schema. This is a low-level method; use keyword
         arguments to the Schema constructor to create the fields instead.
         
         :param name: The name of the field.
-        :param fieldtype: An instantiated fields.FieldType object, or a FieldType subclass.
-            If you pass an instantiated object, the schema will use that as the field
-            configuration for this field. If you pass a FieldType subclass, the schema
-            will automatically instantiate it with the default constructor.
+        :param fieldtype: An instantiated fields.FieldType object, or a
+            FieldType subclass. If you pass an instantiated object, the schema
+            will use that as the field configuration for this field. If you
+            pass a FieldType subclass, the schema will automatically
+            instantiate it with the default constructor.
         """
         
         if name.startswith("_"):
@@ -412,6 +512,8 @@ class Schema(object):
             return self.name_to_number(id)
     
     def to_name(self, id):
+        """Given a field name or number, returns the field's name.
+        """
         if isinstance(id, int):
             return self.number_to_name(id)
         else:
